from typing import List, Dict
import warnings
from alpaca.trading.client import TradingClient
from . import auth
from . import config
from .asset import Stock


class Position:
    def __init__(self, symbol) -> None:
        APCA_ID, APCA_SECRET = auth.get_api_keys()
        self.client = TradingClient(APCA_ID, APCA_SECRET, paper=config.PAPER_TRADING)
        self.symbol = symbol

    def quantity(self) -> int:
        """returns the number of the assets that is owned"""
        qty: int = 0
        try:
            qty = int(self.client.get_open_position(self.symbol).qty)
        except:
            qty = 0
        return qty

    def market_value(self) -> float:
        """Returns market value of symbol in portfolio"""
        return float(self.client.get_open_position(self.symbol).market_value)

    def to_dict(self) -> dict:
        """Returns a dict of the position"""
        return {
            "symbol": self.symbol,
            "market_value": self.market_value(),
            "quantity": self.quantity(),
        }

    def __str__(self) -> str:
        return f"({self.symbol}, {self.quantity()})"

    def __repr__(self) -> str:
        return f"({self.symbol}, {self.quantity()})"


class Portfolio:
    def __init__(self) -> None:
        APCA_ID, APCA_SECRET = auth.get_api_keys()
        self.trading = TradingClient(APCA_ID, APCA_SECRET, paper=config.PAPER_TRADING)
        self.account = self.trading.get_account()
        self.positions: List[Position] = []
        self.account.portfolio_value

    def total_value(self) -> float:
        return float(self.account.portfolio_value)

    def is_blocked(self) -> bool:
        return self.account.account_blocked

    def buying_power(self) -> float:
        """gets the amount of cash currently available"""
        return float(self.account.buying_power)

    def cash(self) -> float:
        """gets the amount of cash currently available"""
        return float(self.account.cash)

    def equity(self) -> float:
        """returns the amount of equity that users has"""
        return float(self.account.equity)

    def equity_yesterday(self) -> float:
        """returns the amount of equity that was
        available at market close yesterday"""
        return float(self.account.last_equity)

    def equity_change(self):
        """returns the change in equity from yesterday to now"""
        return self.equity() - self.equity_yesterday()

    def all_positions(self) -> List[Position]:
        """Returns all positions of portfolio"""
        positions = self.trading.get_all_positions()
        self.positions = []
        for p in positions:
            self.positions.append(Position(p.symbol))
        return self.positions

    def close_all_positions(self) -> None:
        """WARNING: This closes all your open positions"""
        warnings.warn("Warning: will close all open positions ")
        self.trading.close_all_positions(cancel_orders=True)

    def position_in(self, symbol: str) -> Position:
        return Position(symbol)

    def get_allocation(self, symbol: str) -> int:
        return Position(symbol).quantity()

    def cancel_all_orders(self) -> None:
        self.trading.cancel_orders()

    def sell_list(self, symbols: List[str]) -> None:
        """takes a list of Stocks and sells all stocks in that list"""
        for symbol in symbols:
            q = self.position_in(symbol).quantity()
            if q == 0:
                continue
            stk = Stock(symbol=symbol)
            stk.sell(q)

    def owned_stock_qty(self, symbol: str) -> int:  # maby shuld be in stock.Stock
        """returns quantity of owned of a stock Stock (obj)"""
        assert isinstance(symbol, str), "symbol needs to be string"
        return Position(symbol).quantity()

    def owned_stocks_qty(self) -> Dict[str, int]:
        positions = self.trading.get_all_positions()
        result = {}
        for p in positions:
            result[p.symbol] = Position(p.symbol).quantity()
        return result

    def owned_stocks(self) -> List[Stock]:
        """returns a list of owned stocks"""
        return [Stock(p.symbol) for p in self.all_positions()]

    def __repr__(self):
        return f"portfolio({self.equity()})"

    def __str__(self):
        return f"{self.all_positions()}"



from typing import List
from datetime import datetime
import logging
import warnings
import alpaca
from alpaca.data import CryptoHistoricalDataClient, StockHistoricalDataClient
from alpaca.data.requests import StockLatestQuoteRequest
from alpaca.data.requests import CryptoLatestQuoteRequest
from alpaca.data.requests import CryptoBarsRequest, StockBarsRequest
from alpaca.data.timeframe import TimeFrame
from alpaca.trading.requests import MarketOrderRequest
from alpaca.trading.enums import OrderSide, TimeInForce
from alpaca.data.live import StockDataStream
from alpaca.trading.client import TradingClient
from alpaca.trading.requests import LimitOrderRequest
from alpaca.data.models import Bar

import pandas as pd

from . import auth
from . import config
from . import util
from . import log


class Asset:
    def __init__(self, symbol: str) -> None:
        """Interface class"""
        self.symbol = symbol

    def historical_data_df(
        self, start_date: datetime, end_date: datetime
    ) -> pd.DataFrame:
        raise NotImplementedError

    def price(self) -> float:
        raise NotImplementedError

    def __str__(self) -> str:
        return self.symbol

    def __repr__(self) -> str:
        return self.symbol

    def __eq__(self, other):
        if not isinstance(other, Asset):
            raise ValueError
        return self.symbol == other.symbol

    def __ne__(self, other):
        return not self.__eq__(other)


class Stock(Asset):
    def __init__(self, symbol: str) -> None:
        """Exchange for trading stocks"""
        APCA_ID, APCA_SECRET = auth.get_api_keys()
        self.live_client = StockDataStream(APCA_ID, APCA_SECRET)
        self.history = StockHistoricalDataClient(APCA_ID, APCA_SECRET)
        self.trade = TradingClient(APCA_ID, APCA_SECRET, paper=config.PAPER_TRADING)
        self.symbol = symbol

    def price(self) -> float:
        """gets the asking price of the symbol"""
        perms = StockLatestQuoteRequest(symbol_or_symbols=self.symbol)
        return float(self.history.get_stock_latest_quote(perms)[self.symbol].ask_price)

    def live_data(self, async_function_to_resolve_to, run: bool = True) -> None:
        self.live_client.subscribe_quotes(async_function_to_resolve_to, self.symbol)
        if run:
            self.live_client.run()

    def _get_bars(self, start_date: datetime, end_date: datetime):
        """returns aplc bars from the given dates"""
        params = StockBarsRequest(
            symbol_or_symbols=self.symbol,
            timeframe=TimeFrame.Day,
            start=start_date,
            end=end_date,
            adjustment="all",
        )
        return self.history.get_stock_bars(params)

    def historical_data_df(
        self, start_date: datetime, end_date: datetime
    ) -> pd.DataFrame:
        """takes two dates, and returns a data frame with bars from the given dates"""
        data = self._get_bars(start_date, end_date).df
        data = data.reset_index(level="symbol")
        data["timestamp"] = pd.to_datetime(data.index.get_level_values("timestamp"))
        data.set_index("timestamp", inplace=True)
        return data

    def historical_data(self, start_date: datetime, end_date: datetime) -> List[dict]:
        """takes two dates, and returns a list of dicts with bars from the given dates"""
        return self._get_bars(start_date, end_date).dict()[self.symbol]

    def buy(self, qty: float) -> None:
        """Buy the asset,
        qty is the number of the asset that you buy"""
        market_order = MarketOrderRequest(
            symbol=self.symbol,
            qty=qty,
            side=OrderSide.BUY,
            time_in_force=TimeInForce.DAY,
        )
        if config.IS_LOGGING:
            log.log("BUY", self.symbol, qty)
        self.trade.submit_order(market_order)

    def sell(self, qty: float) -> None:
        """Sell the asset,
        qty is the number of the asset that you sell"""
        market_order = MarketOrderRequest(
            symbol=self.symbol,
            qty=qty,
            side=OrderSide.SELL,
            time_in_force=TimeInForce.DAY,
        )
        logging.info(f"sell:{self.symbol}, qty:{qty}")
        if config.IS_LOGGING:
            log.log("SELL", self.symbol, qty)
        self.trade.submit_order(market_order)

    def buy_at(self, qty: int, price: float) -> None:
        """Buy the asset at a given price,
        qty is the number of the asset that you buy"""
        limit_order_data = LimitOrderRequest(
            symbol=self.symbol,
            limit_price=price,
            notional=qty,
            side=OrderSide.BUY,
            time_in_force=TimeInForce.FOK,
        )
        if config.IS_LOGGING:
            log.log("BUY", self.symbol, qty)
        self.trade.submit_order(order_data=limit_order_data)

    def sell_at(self, qty: int, price: float) -> None:
        """Sell the asset at a given price,
        qty is the number of the asset that you sell"""
        limit_order_data = LimitOrderRequest(
            symbol=self.symbol,
            limit_price=price,
            notional=qty,
            side=OrderSide.SELL,
            time_in_force=TimeInForce.FOK,
        )
        if config.IS_LOGGING:
            log.log("SELL", self.symbol, qty)
        self.trade.submit_order(order_data=limit_order_data)

    def save_historical_data(
        self, file_path, start_date: datetime, end_date: datetime
    ) -> None:
        data = self.historical_data_df(start_date=start_date, end_date=end_date)
        data.to_csv(file_path)

    @classmethod
    def load_historical_data(cls, file_path) -> pd.DataFrame:
        """
        Load in model from pickle file
        usage:
            model = Strategy.load('./model.pkl')
            predictions = model.predict(X_test)
        """
        data = pd.read_csv(file_path)
        data["timestamp"] = pd.to_datetime(data["timestamp"])
        data.set_index("timestamp", inplace=True)
        return data

    def value(self) -> float:  # prev: value_of_stock
        """takes a string sym. Gets and returns the stock value at close"""

        warnings.warn(f"Warning: function is deprecated ({self.value})")

        nr_days = 1
        bars = self.barset(nr_days)
        if bars is None:
            self._value = 0.0
        else:
            self._value = bars[0].c  # get stock at close
        return self._value

    def order(self, qty: int, beh: str) -> float:
        """submit order and is a template for order"""

        warnings.warn(f"Warning: function is deprecated ({self.order})")

        if not self.is_tradable():
            raise Exception(
                f"Sorry, {self.symbol} is currantly not tradable on https://alpaca.markets/"
            )
        order = auth.api().submit_order(
            symbol=self.symbol, qty=qty, side=beh, type="market", time_in_force="gtc"
        )
        return order

    def is_sortable(self) -> bool:
        """checks if stock can be shorted"""
        return bool(auth.api().get_asset(self.symbol).shortable)

    def can_borrow(self) -> bool:
        """check whether the name is currently
        available to short at Alpaca"""
        return auth.api().get_asset(self.symbol).easy_to_borrow

    def barset(self, limit: int):
        """returns barset for stock for time period lim"""
        return alpaca.api().get_bars(self.symbol, TimeFrame.Minute, limit=int(limit))

    def is_tradable(self) -> bool:
        """return if the stock can be traded"""
        return bool(auth.api().get_asset(self.symbol).tradable)

    def position(self):
        """returns position of stock"""

        warnings.warn(f"Warning: function is deprecated ({self.position})")

        pos = auth.api().get_position(self.symbol)
        self._position = util.reformat_position(pos)
        return self._position

    def today_plpc(self) -> float:
        """stock today's profit/loss percent"""
        self._today_plpc = self.position()["unrealized_intraday_plpc"]
        return self._today_plpc

    def plpc(self) -> float:
        """stock sym (str) Unrealized profit/loss percentage"""
        self._plpc = self.position()["unrealized_plpc"]
        return self._plpc

    # Operators

    def __eq__(self, other):
        if isinstance(other, (int, float)):
            return self.price() == other
        return self.price() == other.price

    def __ne__(self, other):
        if isinstance(other, (int, float)):
            return self.price() != other
        return self.price() != other.price

    def __lt__(self, other):
        if isinstance(other, (int, float)):
            return self.price() < other
        return self.price() < other.price

    def __le__(self, other):
        if isinstance(other, (int, float)):
            return self.price() <= other
        return self.price() <= other.price

    def __gt__(self, other):
        if isinstance(other, (int, float)):
            return self.price() > other
        return self.price() > other.price

    def __ge__(self, other):
        if isinstance(other, (int, float)):
            return self.price() >= other
        return self.price() >= other.price

    # Arithmetic Operators

    def __add__(self, other):
        if isinstance(other, (int, float)):
            return self.price() + other
        return self.price() + other.price

    def __radd__(self, other):
        return self.price() + other

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            return self.price() - other
        return self.price() - other.price

    def __rsub__(self, other):
        return self.price() - other

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return self.price() * other
        return self.price() * other.price

    def __rmul__(self, other):
        return self.price() * other

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return self.price() / other
        return self.price() / other.price

    def __rdiv__(self, other):
        return self.price() / other

    def __floordiv__(self, other):
        if isinstance(other, (int, float)):
            return self.price() // other
        return self.price() // other.price

    def __rfloordiv__(self, other):
        return self.price() // other

    # Type Conversion

    def __abs__(self):
        # dose not rely makes sense should not be able to
        # be neg but might be good to have
        return abs(self.price)

    def __int__(self):
        return int(self.price)

    def __float__(self):
        return float(self.price)

    def __round__(self, nDigits):
        return round(self.price, nDigits)


class Cryptocurrency(Asset):
    def __init__(self, symbol: str) -> None:
        """Exchange for trading cryptocurrencies"""
        APCA_ID, APCA_SECRET = auth.get_api_keys()
        self.client = CryptoHistoricalDataClient(APCA_ID, APCA_SECRET)
        self.symbol = symbol
        warnings.warn("WARNING: this is very broken and will be fixed later")

    def _get_bars(self, start_date: datetime, end_date: datetime):
        params = CryptoBarsRequest(
            symbol_or_symbols=[self.symbol],
            timeframe=TimeFrame.Day,
            start=start_date,
            end=end_date,
        )
        return self.client.get_crypto_bars(params)

    def historical_data_df(
        self, start_date: datetime, end_date: datetime
    ) -> pd.DataFrame:
        return self._get_bars(start_date, end_date).df

    def price(self) -> float:
        """gets the asking price of the symbol"""
        perms = CryptoLatestQuoteRequest(symbol_or_symbols=self.symbol)
        return float(self.client.get_crypto_latest_quote(perms)[self.symbol].ask_price)



from typing import List
from . import asset
from . import auth
from . import config
from alpaca.trading.client import TradingClient
from alpaca.trading.requests import GetAssetsRequest
from alpaca.data import CryptoHistoricalDataClient, StockHistoricalDataClient
from alpaca.trading.enums import AssetClass
import warnings


class Exchange:
    def __init__(self) -> None:
        APCA_ID, APCA_SECRET = auth.get_api_keys()
        self.alpc_client = TradingClient(APCA_ID, APCA_SECRET)
        self.alpc_historical = StockHistoricalDataClient(APCA_ID, APCA_SECRET)
        self.stock_cache: List[asset.Stock] = []

    def is_open(self) -> bool:
        """Checks if the exchange is open and able to trade"""
        return auth.api().get_clock().is_open

    def to_assets(self, symbols: List[str]) -> List[asset.Asset]:
        """Takes a list of symbols and returns
        them in a list of cira Assets objects"""
        return [self.to_asset(s) for s in symbols]

    def to_asset(self, symbol: str) -> asset.Asset:
        """Takes a symbols and returns
        it as a cira Assets objects"""
        return asset.Stock(symbol)

    def get_all_stocks(
        self, is_tradeable: bool = True, force_reload: bool = False
    ) -> List[asset.Stock]:
        """Returns a list of all stocks as cira asset,
        objects will be cached, can be turn off in config."""
        if config.USE_CASHING and self.stock_cache != [] and not force_reload:
            return self.stock_cache
        search_params = GetAssetsRequest(asset_class=AssetClass.US_EQUITY)
        alpc_assets = self.alpc_client.get_all_assets(search_params)
        self.stock_cache = [
            asset.Stock(a.symbol) for a in alpc_assets if a.tradable == is_tradeable
        ]
        return self.stock_cache

    def calendar(self, start="2018-12-01", end="2018-12-01"):
        self._calendar = (
            auth.api().get_calendar(start=start, end=end)[0].__dict__["_raw"]
        )
        return self._calendar

    def assets_raw(self):
        """(legacy, should not be used)
        returns a list of all avilabel stocks in exchanges list"""
        all_assets = []
        active_assets = auth.api().list_assets(status="active")
        for exchange in self.exchanges:
            all_assets += [a for a in active_assets if a.exchange == exchange]
        self._assets = all_assets
        return self._assets

    def symbols(self):
        """returns a list of all symbols"""
        self._symbols = []
        for asset in self.assets_raw():
            self._symbols.append(asset.__dict__["_raw"]["symbol"])
        return self._symbols


class DemoExchange(Exchange):
    def __init__(self) -> None:
        """uses crypto client, so no need for keys, has limited usage"""
        warnings.warn("Warning: demo exchange has limited usage ")
        self.alpc_historical = CryptoHistoricalDataClient()



import logging

"""
 This files keeps the varibale that is used 
 for configuration cross cira. 
"""

# logging
IS_LOGGING = False
LOG_FILE = "./cira-log.csv"
LOGGING_LEVEL = logging.WARNING

# debugging
DEBUG = False

# paper trading
PAPER_TRADING = True

# data that will not very often change is cached
USE_CASHING = True

# Backtesting
FEE_RATE = 0.004



import auth
from alpaca.trading.client import TradingClient
from .config import PAPER_TRADING


def get_trading_client() -> TradingClient:
    """get the alpaca-sdk python trading class
    obj initalized with set config"""
    apca_id, apca_key = auth.get_api_keys()
    assert apca_id != "" and apca_key != "", "The keys for alpaca are not set "
    return TradingClient(apca_id, apca_key, paper=PAPER_TRADING)



"""
Cira

A simpler libray for alpaca-trade-api from Alpaca Markets.
"""

from . import auth
from . import config
from . import util
from . import log

from .exchange import Exchange, DemoExchange
from .asset import Asset, Stock, Cryptocurrency
from .portfolio import Portfolio, Position

import alpaca

__version__ = "3.0.0"
__author__ = "Axel Gard"
__credits__ = "alpaca.markets"



import csv
import time
import logging
from . import config

"""
This functions is logging trades
"""


def log(action: str, sym: str, qty: int) -> None:
    """writes log data to file"""
    time_ = time.strftime("%Y-%m-%d %H:%M", time.gmtime())
    log_data = [action, sym, qty, time_]
    with open(config.LOG_FILE, "a") as file:
        writer = csv.writer(file)
        writer.writerow(log_data)



def reformat_position(position):
    """reformat position to be float values"""
    raw_position = vars(position)["_raw"]
    position_dict = {}
    for key, value in raw_position.items():
        try:
            if isinstance(value, str):
                if "." in value:
                    position_dict[key] = float(value)
                else:
                    position_dict[key] = int(value)
        except ValueError:
            continue
    return position_dict


def bars_to_dict(bars):
    result = []
    for bar in bars:
        result.append(
            {
                "open": float(bar.open),
                "high": float(bar.high),
                "low": float(bar.low),
                "close": float(bar.close),
                "time": bar.timestamp.strftime("%Y-%m-%d, %H:%M:%S"),
                "time_zone": str(bar.timestamp.tzinfo),
                "volume": float(bar.volume),
            }
        )
    return result


def date_to_days_back(date: str):
    pass



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="cira",
    version="3.0.0",
    description="A simpler library for the alapaca trade api",
    url="https://github.com/AxelGard/cira",
    author="Axel Gard",
    author_email="axel.gard@tutanota.com",
    license="MIT",
    packages=["cira"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "alpaca-py==0.5.4",
        "alpaca-trade-api==2.3.0",
        "schedule==1.2.0",
    ],
    extras_requires={"dev": ["pytest"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Office/Business :: Financial",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
)



import json
import alpaca_trade_api as tradeapi
import os

from alpaca.data import StockHistoricalDataClient, StockLatestQuoteRequest

from .portfolio import Portfolio


"""
This function let's you interact 
with the Alpaca trade API 
"""

KEY_FILE = ""  # user key file path
APCA_API_KEY_ID = ""
APCA_API_SECRET_KEY = ""


def get_api_keys():
    global KEY_FILE
    global APCA_API_KEY_ID
    global APCA_API_SECRET_KEY

    if "APCA_ID" in os.environ:
        APCA_ID = os.environ["APCA_ID"]
        APCA_KEY = os.environ["APCA_KEY"]
    elif KEY_FILE:
        auth_header = authentication_header()
        APCA_ID = str(auth_header["APCA-API-KEY-ID"])
        APCA_KEY = str(auth_header["APCA-API-SECRET-KEY"])
    else:
        APCA_ID = APCA_API_KEY_ID
        APCA_KEY = APCA_API_SECRET_KEY

    if not APCA_ID or not APCA_KEY:
        url = "https://github.com/AxelGard/cira/wiki/Storing-the-Alpaca-API-key"
        raise ValueError("Alpaca market keys were not given, " + url)
    return APCA_ID, APCA_KEY


def check_keys() -> bool:
    try:
        APCA_ID, APCA_KEY = get_api_keys()
        stock_client = StockHistoricalDataClient(APCA_ID, APCA_KEY)
        perms = StockLatestQuoteRequest(symbol_or_symbols="SPY")
        ask = float(stock_client.get_stock_latest_quote(perms)["SPY"].ask_price)

        return True
    except:
        return False


def authentication_header():
    """get's key and returns key in json format"""
    global KEY_FILE
    with open(KEY_FILE, "r") as file:
        header = json.load(file)
    return header


def api(version="v2"):
    """returns object for api"""
    APCA_ID, APCA_KEY = get_api_keys()
    # Open the API connection
    api = tradeapi.REST(APCA_ID, APCA_KEY, "https://paper-api.alpaca.markets", version)
    # Get account info
    api.get_account()
    return api
